/*

Firestorm
by Brent P. Newhall <brent@stun-software.com>

 */

#include "Firestorm.h"




// ---------- Test Bullet




void testBullet( void )
	{
	unsigned int i;
	Bullet *b;
	b = new Bullet( 300.0, 50.0 );
	for( i = 0; i < 20; i++ )
		b->update();
	if( b->location.left != 240.0 )
		{
		printf( "testBullet() failed after 20 updates; left = %f, should be 240.0\n",
		        b->location.left );
		return;
		}
	if( b->location.right != 244.0 )
		{
		printf( "testBullet() failed after 20 updates; right = %f, should be 244.0\n",
		        b->location.right );
		return;
		}
	if( b->color != RED )
		{
		printf( "testBullet() failed - bullet is not red initially.\n" );
		return;
		}
	delete b;
	b = new Bullet( 300.0, 50.0 );
	b->velocity.x = -2.0;
	b->velocity.y = 1.0;
	b->size = BULLET_WAVE;
	b->sync();
	for( i = 0; i < 30; i++ )
		b->update();
	if( b->location.left != 240.0 )
		{
		printf( "testBullet() failed after 30 updates; left = %f, should be 240.0\n",
		        b->location.left );
		return;
		}
	if( b->location.top != 74.0 )
		{
		printf( "testBullet() failed after 30 updates; top = %f, should be 80.0\n",
		        b->location.top );
		return;
		}
	if( b->location.bottom != 116.0 )
		{
		printf( "testBullet() failed after 30 updates; bottom = %f, should be 113.0\n",
		        b->location.bottom );
		return;
		}
	if( b->heartbeat != 30 )
		{
		printf( "testBullet() failed after 30 updates; heartbeat = %d, should be 30\n",
		        b->heartbeat );
		return;
		}
	delete b;
	}




// ---------- Test Player




void testPlayer( void )
	{
	Player player;
	if( player.getColor() != player.RED )
		{
		printf( "testPlayer() not red initially.  Instead, it's %d\n",
		        player.getColor() );
		return;
		}
	if( player.getLocation().x != 10.0 )
		{
		printf( "testPlayer() not at x = 10 initially.  Instead, it's %f\n",
		        player.getLocation().x );
		return;
		}
	if( player.getLocation().y != 50.0 )
		{
		printf( "testPlayer() not at y = 150 initially.  Instead, it's %f\n",
		        player.getLocation().y );
		return;
		}
	if( player.getBounds().left != 5.0  ||
	    player.getBounds().right != 300.0  ||
	    player.getBounds().top != 5.0  ||
	    player.getBounds().bottom != 150.0 )
		{
		printf( "testPlayer() bounds not at 5,5,300,150 initially.  Instead, it's %f,%f,%f,%f\n",
		        player.getBounds().left, player.getBounds().top,
		        player.getBounds().top, player.getBounds().bottom );
		return;
		}
	player.setColor( player.YELLOW );
	if( player.getColor() != player.YELLOW )
		{
		printf( "testPlayer() not yellow when commanded to yellow.  Instead, it's %d\n",
		        player.getColor() );
		return;
		}
	player.setColor( player.RED );
	if( player.getColor() != player.RED )
		{
		printf( "testPlayer() not red when commanded to red.  Instead, it's %d\n",
		        player.getColor() );
		return;
		}
	player.setColor( 5 );
	if( player.getColor() != player.RED )
		{
		printf( "testPlayer() not still red when commanded to 5.  Instead, it's %d\n",
		        player.getColor() );
		return;
		}
	player.setColor( 0 );
	if( player.getColor() != player.RED )
		{
		printf( "testPlayer() not still red when commanded to 0.  Instead, it's %d\n",
		        player.getColor() );
		return;
		}
	player.move( 5.0, -1.0 );
	if( player.getLocation().x != 15.0 )
		{
		printf( "testPlayer() not at x = 15 after move.  Instead, it's %f\n",
		        player.getLocation().x );
		return;
		}
	if( player.getLocation().y != 49.0 )
		{
		printf( "testPlayer() not at y = 150 after move.  Instead, it's %f\n",
		        player.getLocation().y );
		return;
		}
	player.setBounds( 4.0, 4.0, 350.0, 200.0 );
	if( player.getBounds().left != 4.0  ||
	    player.getBounds().right != 350.0  ||
	    player.getBounds().top != 4.0  ||
	    player.getBounds().bottom != 200.0 )
		{
		printf( "testPlayer() bounds not at 4,4,350,200 after being set.  Instead, it's %f,%f,%f,%f\n",
		        player.getBounds().left, player.getBounds().top,
		        player.getBounds().top, player.getBounds().bottom );
		return;
		}
	player.move( -20.0, -500.0 );
	if( player.getLocation().x != player.getBounds().left )
		{
		printf( "testPlayer() not at x = 0 after move beyond bounds.  Instead, it's %f\n",
		        player.getLocation().x );
		return;
		}
	if( player.getLocation().y != player.getBounds().top )
		{
		printf( "testPlayer() not at y = 0 after move beyond bounds.  Instead, it's %f\n",
		        player.getLocation().y );
		return;
		}
	player.move( 1000.0, 1000.0 );
	if( player.getLocation().x != player.getBounds().right )
		{
		printf( "testPlayer() not at boundary after move beyond bounds.  Instead, it's %f\n",
		        player.getLocation().x );
		return;
		}
	if( player.getLocation().y != player.getBounds().bottom )
		{
		printf( "testPlayer() not at boundary after move beyond bounds.  Instead, it's %f\n",
		        player.getLocation().y );
		return;
		}
	}




// ---------- Test Enemy




void testEnemy( void )
	{
	unsigned int i;
	std::vector<float> movements;
	movements.push_back( 1.0 );
	movements.push_back( -3.0 );
	movements.push_back( 0.0 );
	movements.push_back( 50.0 );
	movements.push_back( -2.0 );
	movements.push_back( -2.0 );
	Enemy enemy;
	enemy.SetPath( &movements );
	enemy.pathTicks = 100;
	enemy.update();
	enemy.update();
	if( enemy.velocity.x != -3.0 )
		{
		printf( "testEnemy() returned x velocity of %f instead of -3.0 in first test\n",
		        enemy.velocity.x );
		return;
		}
	if( enemy.velocity.y != 0.0 )
		{
		printf( "testEnemy() returned y velocity of %f instead of 0.0 in first test\n",
		        enemy.velocity.y );
		return;
		}
	for( i = 0; i < 51; i++ )
		enemy.update();
	if( enemy.velocity.x != -2.0 )
		{
		printf( "testEnemy() returned x velocity of %f instead of -2.0 in second test\n",
		        enemy.velocity.x );
		return;
		}
	if( enemy.velocity.x != -2.0 )
		{
		printf( "testEnemy() returned y velocity of %f instead of -2.0 in second test\n",
		        enemy.velocity.y );
		return;
		}
	enemy.setType( 0 );
	if( enemy.image != NULL )
		{
		printf( "testEnemy() loaded an image for type 0.\n" );
		return;
		}
	enemy.setType( 1 );
	if( enemy.image == NULL  ||  ! enemy.image->IsValid() )
		{
		printf( "testEnemy() was unable to load image for type 1.\n" );
		return;
		}
	enemy.setType( 2 );
	if( ! enemy.image->IsValid() )
		{
		printf( "testEnemy() was unable to load image for type 2.\n" );
		return;
		}
	enemy.setType( 3 );
	enemy.addDamage( 1 );
	if( enemy.getHealth() != 2 )
		{
		printf( "testEnemy has %d health after one hit, instead of 2.\n",
		        enemy.getHealth() );
		}
	}




// ---------- Test Enemy Manager




void testEnemyManager( void )
	{
	EnemyManager enemyMgr;
	status_t status;
	unsigned int i, heartbeats;
	View *drawingView = new View( Rect( 0, 0, 680, 300 ), "testDrawingView", CF_FOLLOW_ALL, WID_WILL_DRAW );
	status = enemyMgr.set( "test_level1" );
	if( status != 0 )
		{
		printf( "testEnemyManager() failed to load on EventManager.set( \"test_level1\" )\n" );
		printf( "Got status %d\n", status );
		return;
		}
	enemyMgr.addBullet( 50, 105, 2, 0, 1, 1 );
	Point playerLocation;
	heartbeats = 75;
	for( i = 0; i < heartbeats; i++ )
		enemyMgr.heartbeat( drawingView, playerLocation );
	if( enemyMgr.numOnscreenEnemies() != 1 )
		{
		printf( "testEnemyManager() has incorrect number of enemies after %d heartbeats.\n", heartbeats );
		printf( "Number of enemies: %d\n", enemyMgr.numOnscreenEnemies() );
		return;
		}
	}




// ---------- Test Level




void testLevel( void )
	{
	status_t status;
	LevelEventQueue *levelEventQueue;
	levelEventQueue = new LevelEventQueue();
	status = levelEventQueue->set( "blarg" );
	if( status != 1 )
		{
		printf( "testLevel() returned non-1 on levelEventQueue->set( \"blarg\" )\n" );
		return;
		}
	status = levelEventQueue->set( "test_level1" );
	if( status != 0 )
		{
		printf( "testLevel() failed to load on levelEventQueue->set( \"test_level1\" )\n" );
		return;
		}
	if( levelEventQueue->events.size() != 10 )
		{
		printf( "testLevel() failed.\n" );
		printf( "levelEventQueue->events.size() == %d; should be 10.\n",
		        levelEventQueue->events.size() );
		return;
		}
	status = levelEventQueue->set( "test_level2" );
	if( status != 0 )
		{
		printf( "testLevel() failed to load on levelEventQueue->set( \"test_level2\" )\n" );
		return;
		}
	if( levelEventQueue->events.size() != 5 )
		{
		printf( "testLevel() failed.\n" );
		printf( "levelEventQueue->events.size() == %d; should be 5.\n",
		        levelEventQueue->events.size() );
		return;
		}
	}



// ---------- Test Window




void testWindow( void )
	{
	status_t status;
// Currently causing a coredump, for some strange reason.
/*	GameWindow *gameWindow = new GameWindow( Rect( 20, 40, 700, 360 ) );
	status = gameWindow->loadLevelSet( "blarg" );
	if( status != 1 )
		{
		printf( "testWindow() returned non-1 on GameWindow->loadLevelSet( \"blarg\" )\n" );
		return;
		}
	status = gameWindow->loadLevelSet( "test_level_set" );
	if( status != 0 )
		{
		printf( "testWindow() failed to load on GameWindow->loadLevelSet( \"level_set\" )\n" );
		return;
		}
	if( gameWindow->levels.size() != 2 )
		{
		printf( "testWindow() failed.  " );
		printf( "GameWindow->levels.size() == %d; should be 2.\n",
		        gameWindow->levels.size() );
		return;
		} */
	}



// ---------- Application




class TestApplication : public Application
	{
	public:
		TestApplication( void );
	};




TestApplication :: TestApplication( void )
		: Application( "application/x-vind.STUN-FirestormTest" )
	{
	testBullet();
	testPlayer();
	testEnemy();
	testEnemyManager();
	testLevel();
	testWindow();
	this->PostMessage( M_QUIT );
	}




// ---------- main()




int main( void )
	{
	TestApplication *thisApp;
	thisApp = new TestApplication();
	thisApp->Run();
	return( 0 );
	}






